{
    "namespace": "suzieq",
    "name": "inventory",
    "type": "record",
    "fields": [
        {
            "name": "sqvers",
            "type": "string",
            "partition": 0,
            "default": "2.0",
            "suppress": true,
            "description": "Schema version, not selectable"
        },
        {
            "name": "name",
            "type": "string",
            "key": 2,
            "display": 2,
            "description": "Name of inventory entry"
        },
        {
            "name": "model",
            "type": "string",
            "key": 3,
            "display": 3,
            "description": "Model of the entry"
        },
        {
            "name": "status",
            "type": "string",
            "display": 4,
            "description": "Status of the entry"
        },
        {
            "name": "serial",
            "type": "string",
            "key": 4,
            "description": "Serial number of entry"
        },
        {
            "name": "partNum",
            "type": "string",
            "key": 5,
            "description": "Part number of entry"
        },
        {
            "name": "vendor",
            "type": "string",
            "display": 5,
            "description": "Manufacturing vendor of entry"
        },
        {
            "name": "type",
            "type": "string",
            "display": 6,
            "description": "Type of entry, such as linecard, Xcvr etc."
        },
	{
	    "name": "partType",
	    "type": "string",
	    "display": 8,
	    "description": "Type specific info such as optics and such"
	},
        {
            "name": "version",
            "type": "string",
            "display": 7,
            "description": "Version number of inventory"
        },
        {
            "name": "hostname",
            "type": "string",
            "key": 1,
            "display": 1,
            "partition": 2,
            "description": "Hostname associated with this entry"
        },
        {
            "name": "namespace",
            "type": "string",
            "key": 0,
            "display": 0,
            "partition": 1,
            "description": "Namespace associated with this entry"
        },
        {
            "name": "timestamp",
            "type": "timestamp",
            "display": 8
        },
        {
            "name": "active",
            "type": "boolean",
            "description": "If this entry is active or deleted"
        }
    ]
}
